/*
- Mllerinnerung -
Im EEPROM werden die Tage abgelegt, an denen eine der 4 Tonnen rausgestellt werden 
muss oder Scahdstoffmobil/Grnschnitt/Weihnachtsbaumabholung ist..
Jeden Abend um 18:00Uhr prft der Prozessor, ob am nchsten Tag eine Tonne raus muss.
Falls ja, lt er eine von 5 Anzeigen rausklappen.

Datenformat im Abfallkalender-Array: Fr jeden Tag ein Byte. Tagesbeginn ist 18:00
  Bit0..4: Gesetztes Bit bedeutet:Anzeige ausklappen
  
Beim Start werden nacheinander einmalig alle Anzeigen ausgeklappt.  
Sollte der Prozessor nochmal starten, klappt er tglich alle Anzeigen raus. 
 
Extremes stromsparen:
  - Uhrenquarz an Timer T2; weckt den Prozessor alle 8s
  - Takt auf 128KHz
  - IRQ-Routine mit handoptimiertem Assemblercode
  
  alles Abschalten, was nicht gebraucht wird:
  - ADC
  - Komparator
  - BrownOut-Detektor
  - Debug-Schnittstelle (JTAG)
  
  Pull-up Widerstnde auf allen Eingabeports einschalten, damit die nicht "flattern"
  

Realisiert mit AtMega324P
    interner 128KHz Oszillator

Pinbelegung:
    PA0..4:      Signalausgnge (zum ULN2803) 

    PB3 und PB4: Piezo;

    PD0:         HV_IN
    PD4:         Step-Up-Transistor
 

    TOSC1 + TOSC2: 32768Hz Quartz (Pin28+29)

Timer
  T0: Piezo-Signalgeber
  T1: PWM fr Step-Up
  T2: 32768Hz Quarz

  Fuses:ff d9 e3  (ext/high/low)
        No JTAG
        No div8
        Int. 128KHz, 64ms
    
  Hinweis: mein 644er lie sich mit 128KHz internem Takt nicht flashen,
           aber die Fuses konnte man setzen (mit 30KHz ISP).
           Also geht flashen nun so:
           1) ISP auf 30KHz
           2) Fuses auf 8MHz interner RC umstellen, 
           3) ISP auf 125KHz
           4) Programm flashen
           5) Fuses auf 128KHz interner Takt

*/
#define debug 0

#define F_CPU 128000UL	/* CPU clock in Hertz */

#include <inttypes.h>
#include <util/delay.h> 
#include <avr/eeprom.h> 
#include <avr/io.h>
#include <avr/interrupt.h>


#define U8 uint8_t
#define U16 uint16_t

#define BIT0 1
#define BIT1 2
#define BIT2 4
#define BIT3 8
#define BIT4 16
#define BIT5 32
#define BIT6 64
#define BIT7 128


#define HV_IN (PIND & BIT0)

//--------------------------------------------------------------------------
//Alle 8s wacht der Prozessor auf. 
//Dann dreht er die Uhr weiter und legt sich wieder hin.
//Damit das mglichst schnell geht, ist der Code auf Ausfhrungszeit
//von Hand optimiert, indem mglichst viel in Registern gehalten wird.
//(Ich hatte einfach keinen Bock, den Kram komplett in Assembler zu schreiben)

//Die paar Variablen werden fest Registern zugewiesen




register unsigned char n          asm("r11");
register unsigned char dirtyStart asm("r12");
register unsigned char T2Status   asm("r13");
register unsigned char sTag       asm("r14");  //s8 Startwert am Tagesanfang    
register unsigned char neuerTag   asm("r15");  //Wird am Anfang jedes Tages gesetzt
register unsigned char s8         asm("r16");  //Wird alle 8s runtergezhlt
register unsigned char m20        asm("r17");  //Wird alle 20min runtergezhlt



//Die paar Konstanten auch
register unsigned char _7   asm("r7");
register unsigned char _20  asm("r8");
register unsigned char _72  asm("r9");
register unsigned char _150 asm("r10");

//Bleibt eine 16-Bit Variable fr den Compiler:
U16 day; //Index im Abfallkalender

/* P:  1     Papier
   G   2     Gelber Sack
   HM1 4     Hausmll
   B   8     Biotonne
   S  10     Schadstoffmobil/Tannenbaum/Grnschnitt(Bndel)
*/

#include "kalender.inc"  //Enthaelt fr jeden Tag ein Byte (aktueller Abfallkalender)


//--------------------------------------------------------------------------
U8 getEEPROM(void){
   return eeprom_read_byte((uint8_t*)0);
}
//--------------------------------------------------------------------------
void setEEPROM(void){
   eeprom_write_byte((uint8_t*)0,0);
}


//--------------------------------------------------------------------------
inline void schlafen(void){
  asm("out	0x33, r7");  //SMCR = 7 = _BV(SM1)|_BV(SM0)|_BV(SE);  Power-Save-Mode und SE enable
  asm("SLEEP");
  //Schlft bis T2 berluft (also bis zu 8s)
}

//--------------------------------------------------------------------------
//ISR(TIMER2_OVF_vect){ //8s Interrupt 
void  __attribute__ ((naked)) TIMER2_OVF_vect(void){
  asm("in	r13, 0x3f"  );    //Status sichern

  //Uhr weiterdrehen (Count-Down): 10800 Interrupts am Tag, 450 pro Stunde
  asm("subi	r16, 1");  //  s8--;
  asm("brne	.+12 ");   //  if (!s8){ //20 Minuten sind rum
    asm("mov	r16, r10"); // s8 = _150;
    asm("subi	r17, 1");//    m20--;
    asm("brne	.+6 ");  //    if (!m20){ //Tag ist rum
      asm("mov	r17, r9"); //   m20      = _72;          //Neuer Tag hat 72*20Min
      asm("mov	r15, r9"); //   neuerTag = _72;          //Irgendwas ungleich Null
      asm("mov	r16, r14");//   s8 = sTag;         //Mglichkeit zur Quarzkorrektur      
                       //    }
                       //  }
  asm("out	0x3f, r13");    //Status wiederherstellen
  asm("reti         ");
}

//--------------------------------------------------------------------------
void chargeCapacitor(void){  //Elko auf 24V aufladen
U8 ms=0;
  //Set up the 16-bit timer 1  PWM fr Step-Up  (Mode 15, ca. 3Bit Auflsung, 11,6KHz (128KHz/11)
  OCR1A=8;   // TOP
  OCR1B=1;   // 7.8s ON-Time
  TCNT1=0;
  TCCR1A = _BV(WGM11) |_BV(WGM10) | _BV(COM1B1);  //PWM Signal an OC1B, PD4, Pin18 
  TCCR1B = _BV(WGM13) | _BV(WGM12) 
         | _BV(CS10); //Teilt 128KHz Takt durch 1

  while (!HV_IN){ //Warte bis Endspannung erreicht
    _delay_ms(2);
    ms+=2;
    if (ms>20){  //Pulsabstand periodisch verringern
      ms=0;
      if (OCR1A > 3) OCR1A--;
    }
  }
  TCCR1A = 0; //Timer aus
  TCCR1B = 0; 
  PORTD&=~BIT4;  //Step-Up Transistor aus
}

//--------------------------------------------------------------------------
void beepOn(U8 period){
  //Set up the 8-bit timer 0  (1200Hz Beeper)
  TCCR0A = _BV(WGM01)  //CTC-Mode
        | _BV( COM0A0) //Toggle OC0A on Compare Match
        | _BV( COM0B0);//Toggle OC0B on Compare Match  
  if (
       ((PINB & 0x08) && (PINB & 0x10))
     ||(!(PINB & 0x08) && !(PINB & 0x10))
     ){//Ausgnge sind gleich - schalte einen um (lautes Piepsen)
    TCCR0B = _BV(CS00) | _BV(FOC0A);  // 128KHz; einen Ausgang toggeln
  }else{
    TCCR0B = _BV(CS00);  // Clock Select 001 => Teiler=1 (gibt 128KHz)
  }
  OCR0A = period;  
  OCR0B = period;
  DDRB=BIT3|BIT4; //Port auf Ausgang
}
//--------------------------------------------------------------------------
void beepOff(void){
  TCCR0B = _BV(FOC0A);  // Timer aus, Ausgnge gleichschalten
  TCNT0=0;
  DDRB=0;     //Port auf Eingang und
  PORTB=0xff; //Pull-ups ein 
}
//--------------------------------------------------------------------------
void beep(U8 period){  //period=17: 128K/18/2=3555Hz
  beepOn(period);
  _delay_ms(1200); //Piepsdauer in ms
  beepOff();
}


//--------------------------------------------------------------------------
void eject(U8 channel){
  chargeCapacitor();
  PORTA = (1 << channel);
  beep(17); //includes delay
  PORTA = 0;
}
//--------------------------------------------------------------------------
void stundengong(void){
U8 h,m;

  if (s8) return;  
  m = m20 % 3;
  if (!m){
    m=3;
    while(m--){
      beepOn(12);
      _delay_ms(200);  
      beepOff();
      _delay_ms(800);
    }    
    _delay_ms(2000);

    h= 24 - m20/3;
    if (h>12) h-=12;
    while(h--){
      beepOn(15);
      _delay_ms(80);  
      beepOff();
      _delay_ms(920);
    }    
  }else{
    while(m--){
      beepOn(11);
      _delay_ms(200);  
      beepOff();
      _delay_ms(800);
    }    
  }
}
//--------------------------------------------------------------------------
int main(void){

  //CLKPR = _BV(CLKPS2)|_BV(CLKPS1);  //8MHz / 64 = 128KHz Systemtakt

  cli();
  ADCSRA=0; //A/D Wandler aus
  ACSR=_BV(ACD);  //Analog Comparator disable


  //Watchdog tten
  asm("wdr");// watchdog reset
  MCUSR &= ~(1<<WDRF); // Clear WDRF in MCUSR 
  /* Write logical one to WDCE and WDE */
  /* Keep old prescaler setting to prevent unintentional time-out */
  WDTCSR |= (1<<WDCE) | (1<<WDE);
  WDTCSR = 0; /* Turn off WDT */

  MCUCR = 0x80; //JTD=1: Jtag Debug disable (muss 2x geschrieben werden)
  MCUCR = 0x80; 




  //Port Direction: 1=Output 0=Input
  DDRA=BIT0|BIT1|BIT2|BIT3|BIT4; //PA0..PA4, Pin 40..36
  DDRB=BIT3|BIT4; //Piezo an PB3/PB4 (OC0A/OC0B), Pin 4/5
  DDRC=0;  
  DDRD=BIT4;  //PWM-Transistor

  PORTA=                        BIT5|BIT6|BIT7; //Pull-Ups einschalten
  PORTB=BIT0|BIT1|BIT2|         BIT5|BIT6|BIT7; //Pull-Ups einschalten
  PORTC=0xff; //Pull-Ups einschalten
  PORTD=     BIT1|BIT2|BIT3|    BIT5|BIT6|BIT7; //Pull-Ups einschalten

  //Set up the 8-bit timer 0  (1500Hz Beeper at startup)
  TCCR0A = _BV(WGM01)  //CTC-Mode
        | _BV( COM0A0) //Toggle OC0A on Compare Match
        | _BV( COM0B0);//Toggle OC0B on Compare Match  
  TCCR0B = _BV(CS00) | _BV(FOC0A);  // Clock Select 001 => Teiler=1 (gibt 128KHz)
  OCR0A = 30;
  OCR0B = 30;
  TIMSK0 = 0;

  //Set up the 16-bit timer 1  PWM fr Step-Up
  TCCR1A = 0;
  TCCR1B = 0;
  TIMSK1 = 0;
 
  //Set up the 8-bit timer 2  (32768Hz RealTime Clock)
  //--------------------------------------------------
  TCCR2A =  0;
  TCCR2B =  
            _BV(CS22)  
          | _BV(CS21)  // Clock Select 111 => Teiler=1024 (gibt 32Hz)
          | _BV(CS20) 
          ;
  ASSR = _BV(AS2); //Asyncron Real-Time-Clock; EXCLK=0

  //Timer Interrupt Mask
  TIMSK2 = _BV(TOIE2);  // Timer2 Overflow Interrupt aktivieren
  //Timer luft alle 8 Sekunden ber (32Hz/256)

  //Konstante Register setzen
  _7 =7;
  _20=20;
  _72=72;
  _150=150;



  //Nach Flashen Zeit lassen, den Prozessor abzuklemmen und auf Batterie laufen zu lassen
  _delay_ms(2000); //Tiefer Ton erklingt 
  beep(17);        //Hoher Ton fr 1,2s
  _delay_ms(3000); 
  _delay_ms(3000); 
  //Alle Ausgnge nacheinander feuern
  for (n=0; n<5; n++){
    eject(n);
  }

  dirtyStart=0;
  //Prfen, ob dieses der 1. Start nach dem Flashen ist.
  //Nur beim 1. Start nach dem Programmieren ist alles OK. Sonst stimmt ja die Zeit/der Tag nicht!
  if (0xff != getEEPROM()){ //gelschtes EEPROM enthlt 0xFF
    dirtyStart=1;
  }else{
    setEEPROM(); //Dirty-Flag setzen
  }
  
    
  //Neuer Tag: Startwert des ersten 20Min.Intervalls
  //Mglichkeit zur Quarzkorrektur (8s/Tag je Schritt). 
  sTag=150;  //150*8s = 20 Min
             //149 macht Uhr also 8s/Tag schneller
             //152 macht Uhr also 16s/Tag langsamer

  initClock();
#if debug
    DDRD|=BIT7;
#endif    

  sei(); //Interrupts ein
  
  //-------------------------------------------------------
  //   Hauptschleife
  //-------------------------------------------------------
  while(1){ //Arbeite bis die Batterie rausgenomen wird

    schlafen(); //8s nix tun. T2 weckt via Interrupt.
#if debug
    PIND|=BIT7; //Debug pin togglen
    beepOn(10);
    _delay_ms(2);
    beepOff();
#endif    
    stundengong();  //<--- auskommentieren wenn Uhr getestet

    if (neuerTag){  //In der realen Welt: 18:00Uhr
      neuerTag=0;
      day++;
      if (day > 366) dirtyStart=1; //Neuer Kalender ist fllig

      beep(14); //Hoher Ton fr neuen Tag
      //128K/15/2=4266Hz

      if (dirtyStart){ //
        //Alle Ausgnge nacheinander feuern - Anwender muss selbst nachsehen was dran ist.
        for (n=0; n<5; n++){
          eject(n);
        }
      }else{ //Uhr luft normal. 
        for (n=0; n<5; n++){
          if (abfallkalender[day] & (1<<n)) eject(n);
        }
      }//else dirtyStart
    }//neuer Tag
  }//While 1
}

